require "lib.AutoTable"
Class = require 'lib.hump.class'
Ringbuffer = require 'lib.hump.ringbuffer'

gDebugEnabled = false

function newRingbufferHead()
  local head = AutoTable()
  head.isHead = true
  return head
end

function newRingBuffer()
  return Ringbuffer(newRingbufferHead())
end

function printDebugButtonState(widget)
  if not gDebugEnabled then return end
  local isChecked = widget.properties.isChecked
  local isMouseOver = widget.properties.isMouseOver
  local isMousePressed = widget.properties.isMousePressed
  print("Current button state:")
  print("  : isChecked = " .. tostring(isChecked))
  print("  : isMouseOver = " .. tostring(isMouseOver))
  print("  : isMousePressed = " .. tostring(isMousePressed))
  print("\n")
end


function setProperty(widget, property, value)
  widget.properties[property] = value

  if not gDebugEnabled then return end
  print("setProperty called:")
  print("  : widget = " .. tostring(widget))
  print("  : property = " .. tostring(property))
  print("  : value = " .. tostring(value))
  print("\n")
end

function getProperty(widget, property)
  return widget.properties[property]
end

function toggleBooleanProperty(widget, property)
  local currentValue = widget.properties[property]
  setProperty(widget, property, not currentValue)

  if not gDebugEnabled then return end
  print("toggleProperty called:")
  print("  : widget = " .. tostring(widget))
  print("  : property = " .. tostring(property))
  print("  : value = " .. tostring(value))
  print("\n")
end

function updateButtonImage(widget)
  local isChecked = widget.properties.isChecked
  local isMouseOver = widget.properties.isMouseOver
  local isMousePressed = widget.properties.isMousePressed

  if not isMousePressed and not isMouseOver and not isChecked then
    setProperty(widget, "currentImage", 2)
  end
  if not isMousePressed and isMouseOver and not isChecked then
    setProperty(widget, "currentImage", 3)
  end
  if isMousePressed and isMouseOver and not isChecked then
    setProperty(widget, "currentImage", 4)
  end
  if isMousePressed and isMouseOver and isChecked then
    setProperty(widget, "currentImage", 5)
  end
  if not isMousePressed and isMouseOver and isChecked then
    setProperty(widget, "currentImage", 6)
  end
  if not isMousePressed and not isMouseOver and isChecked then
    setProperty(widget, "currentImage", 7)
  end

  if not gDebugEnabled then return end
  print("Current button state:")
  print("  : isChecked = " .. tostring(isChecked))
  print("  : isMouseOver = " .. tostring(isMouseOver))
  print("  : isMousePressed = " .. tostring(isMousePressed))
  print("\n")
end

function raiseEvent(widget, event, args)
  local method = widget.events[event]
  widget.methods[method](args)

  if not gDebugEnabled then return end
  print("setProperty called:")
  print("  : widget = " .. tostring(widget))
  print("  : event = " .. tostring(event))
  print("  : args = " .. tostring(args))
  print("  : calling " .. widget.name .. ".methods." .. tostring(method))
  print("\n")
end

function checkMouseHit(widget, x, y, button, action)
  local x0 = gToggleWidget.properties.margin.left
  local y0 = gToggleWidget.properties.margin.top
  local x1 = x0 + gToggleWidget.properties.width
  local y1 = y0 + gToggleWidget.properties.height

  local inBounds = false
  if x > x0 and x < x1 and y > y0 and y < y1 then
    inBounds = true
  end

  if action == "pressed" then
    if inBounds == true then
      widget.methods.onMousePressed()
    end
  elseif action == "released" then
    widget.methods.onMouseReleased(inBounds)
  else
    if inBounds == true then
      if widget.properties.isMouseOver == false then
        widget.properties.isMouseOver = true
        raiseEvent(widget, "mouseEntry", nil)
      end
    else
      if widget.properties.isMouseOver == true then
        widget.properties.isMouseOver = false
        raiseEvent(widget, "mouseExit", nil)
      end
    end
  end

  return inBounds
end


gUndoManager = AutoTable()

gWindow = {}
gWindow.frames = newRingBuffer()

gMainFrame = {}
gMainFrame.widgets = newRingBuffer()

gToggleWidget = {}
gToggleWidget.name = "gToggleWidget"
gToggleWidget.properties = AutoTable()
gToggleWidget.events = AutoTable()
gToggleWidget.methods = AutoTable()

gToggleWidget.properties.isChecked = false
gToggleWidget.properties.isMouseOver = false
gToggleWidget.properties.isMousePressed = false
gToggleWidget.properties.content = "Unchecked"
gToggleWidget.properties.width = 117
gToggleWidget.properties.height = 20
gToggleWidget.properties.margin = {top = 118, bottom = 0, left = 69, right = 0}
gToggleWidget.properties.backgroundColor = {r=255, g=0, b=0, a=255}
gToggleWidget.properties.images[1] = love.graphics.newImage("assets/images/gui/button/border_invariant.png")
gToggleWidget.properties.images[2] = love.graphics.newImage("assets/images/gui/button/u_n_u_0.png")
gToggleWidget.properties.images[3] = love.graphics.newImage("assets/images/gui/button/u_h_u_1.png")
gToggleWidget.properties.images[4] = love.graphics.newImage("assets/images/gui/button/d_h_u_2.png")
gToggleWidget.properties.images[5] = love.graphics.newImage("assets/images/gui/button/d_h_c_3.png")
gToggleWidget.properties.images[6] = love.graphics.newImage("assets/images/gui/button/u_h_c_4.png")
gToggleWidget.properties.images[7] = love.graphics.newImage("assets/images/gui/button/u_n_c_5.png")
gToggleWidget.properties.currentImage = 2

gToggleWidget.events.checked = "onChecked"
gToggleWidget.events.unchecked = "onUnchecked"
gToggleWidget.events.mousePressed = "onMousePressed"
gToggleWidget.events.mouseReleased = "onMouseReleased"
gToggleWidget.events.mouseEntry = "onMouseEntry"
gToggleWidget.events.mouseExit = "onMouseExit"


gToggleWidget.methods.onMethodCall = function(e)
  printDebugButtonState(gToggleWidget)
end

gToggleWidget.methods.onMethodCallFinish = function(e)
  updateButtonImage(gToggleWidget)
end

gToggleWidget.methods.onChecked = function(e)
    gToggleWidget.methods.onMethodCall()
    setProperty(gToggleWidget, "content", "stop simulation")
    gToggleWidget.methods.onMethodCallFinish()

    if not gDebugEnabled then return end
    print("gToggleWidget.methods.onChecked called")
    print("\n")
  end

gToggleWidget.methods.onUnchecked = function(e)
    gToggleWidget.methods.onMethodCall()
    setProperty(gToggleWidget, "content", "start simulation")
    gToggleWidget.methods.onMethodCallFinish()

    if not gDebugEnabled then return end
    print("gToggleWidget.methods.onUnchecked called")
    print("\n")
  end

gToggleWidget.methods.onMousePressed = function(e)
    gToggleWidget.methods.onMethodCall()
    setProperty(gToggleWidget, "isMousePressed", true)
    gToggleWidget.methods.onMethodCallFinish()

    if not gDebugEnabled then return end
    print("gToggleWidget.methods.onMousePressed called")
    print("\n")
  end

gToggleWidget.methods.onMouseReleased = function(inBounds)
    gToggleWidget.methods.onMethodCall()

    if true == inBounds then
      toggleBooleanProperty(gToggleWidget, "isChecked")

      local isChecked = getProperty(gToggleWidget, "isChecked")

      if true == isChecked then
        raiseEvent(gToggleWidget, "checked", nil)
      elseif false == isChecked then
        raiseEvent(gToggleWidget, "unchecked", nil)
      end
    end

    setProperty(gToggleWidget, "isMousePressed", false)

    print(love.mouse.getX())

    gToggleWidget.methods.onMethodCallFinish()

    if not gDebugEnabled then return end
    print("gToggleWidget.methods.onMouseReleased called")
  end

gToggleWidget.methods.onMouseEntry = function(e)
    gToggleWidget.methods.onMethodCall()

    setProperty(gToggleWidget, "isMouseOver", true)

    gToggleWidget.methods.onMethodCallFinish()

    if not gDebugEnabled then return end
    print("gToggleWidget.methods.onMouseEntry called")
  end

gToggleWidget.methods.onMouseExit = function(e)
    gToggleWidget.methods.onMethodCall()
    setProperty(gToggleWidget, "isMouseOver", false)
    gToggleWidget.methods.onMethodCallFinish()

    if not gDebugEnabled then return end
    print("gToggleWidget.methods.onMouseExit called")
  end


gToggleWidget.paint = function(self)
  local r, g, b, a = love.graphics.getColor()

  r1 = gToggleWidget.properties.backgroundColor.r
  g1 = gToggleWidget.properties.backgroundColor.g
  b1 = gToggleWidget.properties.backgroundColor.b
  a1 = gToggleWidget.properties.backgroundColor.a

  local x0 = gToggleWidget.properties.margin.left
  local y0 = gToggleWidget.properties.margin.top
  local x1 = x0 + gToggleWidget.properties.width
  local y1 = y0 + gToggleWidget.properties.height

  love.graphics.setColor(255, 255, 255, 255)
  -- love.graphics.polygon("fill", x0, y0, x1, y0, x1, y1, x0, y1)
  love.graphics.draw(gToggleWidget.properties.images[1], x0, y0)
  local currentImage = gToggleWidget.properties.currentImage
  love.graphics.draw(gToggleWidget.properties.images[currentImage], x0, y0)

  local wrapLimit = gToggleWidget.properties.width
  local fontSize = 12
  local textHeight = (y1 - y0) / 2 + y0 - (fontSize * (1/2))

  -- local buttonText = gToggleWidget.properties.content
  local buttonText = string.format("%d, %d", love.mouse.getPosition())

  love.graphics.setColor(0, 0, 0, 100)
  love.graphics.printf(buttonText, x0, textHeight+1, wrapLimit, "center")

  if true == gToggleWidget.properties.isChecked then
    love.graphics.setColor(0, 0, 0, 255)
  else
    love.graphics.setColor(255, 255, 255, 255)
  end
  love.graphics.printf(buttonText, x0, textHeight, wrapLimit, "center")

  love.graphics.setColor(r, g, b, a)
end


function love.load()
  love.graphics.setMode( 256, 256, false, false, 0 )
  -- love.graphics.setFont(love.graphics.newFont("assets/seg.ttf", 12))
  love.graphics.setFont(love.graphics.newFont(11))
  love.graphics.setBackgroundColor(47, 47, 47, 255)
end

function love.draw()
  gToggleWidget.paint(gToggleWidget)
end

function love.update(dt)
  local mx, my = love.mouse.getPosition()
  checkMouseHit(gToggleWidget, mx, my, button)
end

function love.keypressed(key)
  if key == "escape" then
    love.event.push("quit")
  end
end

function love.keyreleased(key)
end

function love.mousepressed(x, y, button)
  checkMouseHit(gToggleWidget, x, y, button, "pressed")
end

function love.mousereleased(x, y, button)
  checkMouseHit(gToggleWidget, x, y, button, "released")
end

function love.focus(f)
end

function love.quit()
end
